import { Config } from '../core/Config.js'
import type { FnEqual, FnUnequal } from '../interfaces/relational.js'

const DBL_EPSILON = Number.EPSILON || 2.2204460492503130808472633361816E-16

export const equal =
   (dep: {
      config: Config
   }): FnEqual<number> => (x, y) => {
      const eps = dep.config.epsilon
      if (eps === null || eps === undefined) return x === y
      if (x === y) return true
      if (isNaN(x) || isNaN(y)) return false

      if (isFinite(x) && isFinite(y)) {
         const diff = Math.abs(x - y)
         if (diff < DBL_EPSILON) return true
         return diff <= Math.max(Math.abs(x), Math.abs(y)) * eps
      }

      return false
   }

export const unequal = (dep: {
   equal: FnEqual<number>
}): FnUnequal<number> =>
   (x, y) => !dep.equal(x, y)
